from .const import Bound, inf
from .interval import Interval, open, closed, openclosed, closedopen, empty, singleton
from .func import iterate
from .io import from_string, to_string, from_data, to_data
from .dict import IntervalDict


__all__ = [
    'inf', 'CLOSED', 'OPEN',
    'Interval',
    'open', 'closed', 'openclosed', 'closedopen', 'singleton', 'empty',
    'iterate',
    'from_string', 'to_string', 'from_data', 'to_data',
    'IntervalDict',
]

CLOSED = Bound.CLOSED
OPEN = Bound.OPEN



from .const import Bound
from .interval import Interval, singleton

from collections.abc import MutableMapping


class IntervalDict(MutableMapping):
    """
    An IntervalDict is a dict-like data structure that maps from intervals to data,
    where keys can be single values or Interval instances.

    When Interval instances are used as keys, its behaviour merely corresponds to
    range queries and it returns IntervalDict instances corresponding to the
    subset of values covered by the interval. When single values are used as keys,
    its behaviour corresponds to built-in dict.

    KeyError are only raised when values are used as keys. When intervals are used,
    and if there is no matching value, an empty IntervalDict is returned instead.

    Note that this class does not aim to have the best performance, but is
    provided mainly for convenience.
    """

    __slots__ = ('_items', )

    def __init__(self, mapping_or_iterable=None):
        """
        Return a new IntervalDict.

        If no argument is given, an empty IntervalDict is created. If an argument
        is given, and is a mapping object (e.g., another IntervalDict), an
        new IntervalDict with the same key-value pairs is created.

        If an iterable is provided, it has to be a list of (key, value) pairs.

        :param mapping_or_iterable: optional mapping or iterable.
        """
        self._items = list()  # List of (interval, value) pairs

        if mapping_or_iterable is not None:
            self.update(mapping_or_iterable)

    def clear(self):
        """
        Remove all items from the IntervalDict.
        """
        self._items.clear()

    def copy(self):
        """
        Return a shallow copy.

        :return: a shallow copy.
        """
        return IntervalDict(self)

    def get(self, key, default=None):
        """
        Return the values associated to given key.

        If the key is a single value, it returns a single value (if it exists) or
        the default value. If the key is an Interval, it returns a new IntervalDict
        restricted to given interval. In that case, the default value is used to
        "fill the gaps" (if any) w.r.t. given key.

        :param key: a single value or an Interval instance.
        :param default: default value (default to None).
        :return: an IntervalDict, or a single value if key is not an Interval.
        """
        if isinstance(key, Interval):
            d = self[key]
            d[key - d.domain()] = default
            return d
        else:
            try:
                return self[key]
            except KeyError:
                return default

    def find(self, value):
        """
        Return a (possibly empty) Interval i such that self[i] = value, and
        self[~i] != value.

        :param value: value to look for.
        :return: an Interval instance.
        """
        return Interval(*(i for i, v in self._items if v == value))

    def items(self):
        """
        Return a sorted list of (Interval, value) pairs.

        :return: a sorted list of 2-uples.
        """
        def func(i):
            return (i[0].lower, i[0].left is Bound.CLOSED, i[0].upper, i[0].right is Bound.OPEN)

        return sorted(self._items, key=func)

    def keys(self):
        """
        Return the list of underlying Interval instances.

        :return: a list of intervals.
        """
        return [i for i, v in self.items()]

    def values(self):
        """
        Return the list of values.

        :return: a list of values.
        """
        return [v for i, v in self.items()]

    def domain(self):
        """
        Return an Interval corresponding to the domain of this IntervalDict.

        :return: an Interval.
        """
        return Interval(*(i for i, v in self._items))

    def pop(self, key, default=None):
        """
        Return and remove given key.

        This method combines self[key] and del self[key]. If a default value
        is provided and is not None, it uses self.get(key, default) instead of
        self[key].

        :param key: a single value or an Interval instance.
        :param default: optional default value.
        :return: an IntervalDict, or a single value if key is not an Interval.
        """
        if default is None:
            value = self[key]
            del self[key]
            return value
        else:
            value = self.get(key, default)
            try:
                del self[key]
            except KeyError:
                pass
            return value

    def popitem(self):
        """
        Pop an arbitrary existing key.

        :return: an IntervalDict
        """
        try:
            return self.pop(self._items[-1][0])
        except IndexError:
            raise KeyError('Instance is empty.')

    def setdefault(self, key, default=None):
        """
        Return given key. If it does not exist, set its value to default and
        return it.

        :param key: a single value or an Interval instance.
        :param default: default value (default to None).
        :return: an IntervalDict, or a single value if key is not an Interval.
        """
        if isinstance(key, Interval):
            value = self.get(key, default)
            self.update(value)
            return value
        else:
            try:
                return self[key]
            except KeyError:
                self[key] = default
                return default

    def update(self, mapping_or_iterable):
        """
        Update current IntervalDict with provided values.

        If a mapping is provided, it must map Interval instances to values (e.g., another
        IntervalDict). If an iterable is provided, it must consist of a list of
        (key, value) pairs.

        :param mapping_or_iterable: mapping or iterable.
        """
        if hasattr(mapping_or_iterable, 'items'):
            data = mapping_or_iterable.items()
        else:
            data = mapping_or_iterable

        for i, v in data:
            i = singleton(i) if not isinstance(i, Interval) else i
            self[i] = v

    def __getitem__(self, key):
        if isinstance(key, Interval):
            items = []
            for i, v in self._items:
                intersection = key & i
                if not intersection.empty:
                    items.append((intersection, v))
            return IntervalDict(items)
        else:
            for i, v in self._items:
                if key in i:
                    return v
            raise KeyError(key)

    def combine(self, other, how):
        """
        Return a new IntervalDict that combines the values from current and
        provided ones.

        If d = d1.combine(d2, f), then d contains (1) all values from d1 whose
        keys do not intersect the ones of d2, (2) all values from d2 whose keys
        do not intersect the ones of d1, and (3) f(x, y) for x in d1, y in d2 for
        intersecting keys.

        :param other: another IntervalDict instance.
        :param how: a function of two parameters that combines values.
        :return: a new IntervalDict instance.
        """
        new_items = []

        dom1, dom2 = self.domain(), other.domain()

        new_items.extend(self[dom1 - dom2].items())
        new_items.extend(other[dom2 - dom1].items())

        intersection = dom1 & dom2
        d1, d2 = self[intersection], other[intersection]

        for i1, v1 in d1.items():
            for i2, v2 in d2.items():
                if i1.overlaps(i2):
                    i = i1 & i2
                    v = how(v1, v2)
                    new_items.append((i, v))

        return IntervalDict(new_items)

    def __setitem__(self, key, value):
        interval = key if isinstance(key, Interval) else singleton(key)

        if interval.empty:
            return

        new_items = []
        found = False
        for i, v in self._items:
            if value == v:
                found = True
                new_items.append((i | interval, v))
            elif i.overlaps(interval):
                remaining = i - interval
                if not remaining.empty:
                    new_items.append((remaining, v))
            else:
                new_items.append((i, v))

        if not found:
            new_items.append((interval, value))

        self._items = new_items

    def __delitem__(self, key):
        interval = key if isinstance(key, Interval) else singleton(key)

        if interval.empty:
            return

        new_items = []
        found = False
        for i, v in self._items:
            if i.overlaps(interval):
                found = True
                remaining = i - interval
                if not remaining.empty:
                    new_items.append((remaining, v))
            else:
                new_items.append((i, v))

        self._items = new_items

        if not found and not isinstance(key, Interval):
            raise KeyError(key)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self._items)

    def __contains__(self, key):
        return key in self.domain()

    def __repr__(self):
        return '{}{}{}'.format(
            '{',
            ', '.join('{!r}: {!r}'.format(i, v) for i, v in self.items()),
            '}',
        )

    def __eq__(self, other):
        if isinstance(other, IntervalDict):
            return self.items() == other.items()
        else:
            return NotImplemented



import pytest

from portion.const import inf, _PInf, _NInf, Bound
from portion import CLOSED, OPEN


class TestBounds:
    def test_comparisons(self):
        assert CLOSED == Bound.CLOSED
        assert OPEN == Bound.OPEN
        assert CLOSED != OPEN
        assert OPEN != CLOSED

    def test_non_Boolean(self):
        with pytest.raises(ValueError):
            assert not OPEN

        with pytest.raises(ValueError):
            assert not CLOSED

    def test_value(self):
        assert Bound(True) == Bound.CLOSED
        assert Bound(False) == Bound.OPEN

        assert Bound.CLOSED.value == True
        assert Bound.OPEN.value == False


class TestInfinities:
    def test_pinf_is_greater(self):
        assert not (inf > inf)
        assert inf > -inf
        assert inf > 0
        assert inf > 'a'
        assert inf > []

        assert inf >= inf
        assert inf >= -inf
        assert inf >= 0
        assert inf >= 'a'
        assert inf >= []

    def test_ninf_is_lower(self):
        assert not (-inf < -inf)
        assert -inf < inf
        assert -inf < 0
        assert -inf < 'a'
        assert -inf < []

        assert -inf <= -inf
        assert -inf <= inf
        assert -inf <= 0
        assert -inf <= 'a'
        assert -inf <= []

    def test_equalities(self):
        assert inf != -inf
        assert inf == inf
        assert -inf == -inf

    def test_infinities_are_singletons(self):
        assert _PInf() is _PInf()
        assert inf is _PInf()

        assert _NInf() is _NInf()
        assert -inf is _NInf()

        assert -(-inf) is inf

    def test_infinities_are_hashable(self):
        assert hash(inf) is not None
        assert hash(-inf) is not None
        assert hash(inf) != hash(-inf)



import pytest

import portion as P


class TestIntervalDict:
    def test_with_single_values(self):
        d = P.IntervalDict()

        # Single value
        d[P.closed(0, 2)] = 0
        assert len(d) == 1
        assert d[2] == 0
        assert d.get(2) == 0
        with pytest.raises(KeyError):
            d[3]
        assert d.get(3) is None

    def test_with_intervals(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        assert d[P.open(-P.inf, P.inf)].items() == [(P.closed(0, 2), 0)]
        assert d[P.closed(0, 2)].items() == [(P.closed(0, 2), 0)]
        assert d[P.closed(-1, 0)].items() == [(P.singleton(0), 0)]
        assert d[P.closed(-2, -1)].items() == []
        assert d.get(P.closed(0, 2)).items() == [(P.closed(0, 2), 0)]
        assert d.get(P.closed(-2, -1)).items() == [(P.closed(-2, -1), None)]
        assert d.get(P.closed(-1, 0)).items() == [(P.closedopen(-1, 0), None), (P.singleton(0), 0)]

        d[P.closed(1, 3)] = 1
        assert d.items() == [(P.closedopen(0, 1), 0), (P.closed(1, 3), 1)]
        assert len(d) == 2
        assert d[0] == 0
        assert d.get(0, -1) == 0
        assert d[1] == 1
        assert d.get(1, -1) == 1
        assert d[3] == 1
        assert d.get(3, -1) == 1
        with pytest.raises(KeyError):
            d[4]
        assert d.get(4, -1) == -1

    def test_set(self):
        # Set values
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        d[3] = 2
        assert d.items() == [(P.closed(0, 2), 0), (P.singleton(3), 2)]
        d[3] = 3
        assert d.items() == [(P.closed(0, 2), 0), (P.singleton(3), 3)]
        d[P.closed(0, 2)] = 1
        assert d.items() == [(P.closed(0, 2), 1), (P.singleton(3), 3)]
        d[P.closed(-1, 1)] = 2
        assert d.items() == [(P.closed(-1, 1), 2), (P.openclosed(1, 2), 1), (P.singleton(3), 3)]

        d = P.IntervalDict([(P.closed(0, 2), 0)])
        d[P.closed(-1, 4)] = 1
        assert d.items() == [(P.closed(-1, 4), 1)]
        d[P.closed(5, 6)] = 1
        assert d.items() == [(P.closed(-1, 4) | P.closed(5, 6), 1)]

    def test_delete_value(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        del d[1]
        assert d.get(1, None) is None

    def test_delete_missing_value(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        with pytest.raises(KeyError):
            del d[3]

        del d[1]
        with pytest.raises(KeyError):
            d[1]

    def test_delete_interval(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        del d[P.closed(-1, 1)]
        assert d.items() == [(P.openclosed(1, 2), 0)]

    def test_delete_interval_out_of_bound(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        del d[P.closed(-10, -9)]
        assert d.items() == [(P.closed(0, 2), 0)]

    def test_delete_empty_interval(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        del d[P.empty()]
        assert d.items() == [(P.closed(0, 2), 0)]

    def test_setdefault_with_values(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        assert d.setdefault(-1, default=0) == 0
        assert d[-1] == 0
        assert d.setdefault(0, default=1) == 0
        assert d[0] == 0

    def test_setdefault_with_intervals(self):
        d = P.IntervalDict([(P.closed(0, 2), 0)])
        t = d.setdefault(P.closed(-2, -1), -1)
        assert t.items() == [(P.closed(-2, -1), -1)]
        assert d.items() == [(P.closed(-2, -1), -1), (P.closed(0, 2), 0)]

        d = P.IntervalDict([(P.closed(0, 2), 0)])
        t = d.setdefault(P.closed(-1, 1), 2)
        assert t.items() == [(P.closedopen(-1, 0), 2), (P.closed(0, 1), 0)]
        assert d.items() == [(P.closedopen(-1, 0), 2), (P.closed(0, 2), 0)]

    def test_iterators(self):
        d = P.IntervalDict([(P.closedopen(0, 1), 0), (P.closedopen(1, 3), 1), (P.singleton(3), 2)])

        assert d.keys() == [P.closedopen(0, 1), P.closedopen(1, 3), P.singleton(3)]
        assert d.domain() == P.closed(0, 3)
        assert d.values() == [0, 1, 2]
        assert d.items() == list(zip(d.keys(), d.values()))
        assert list(d) == d.keys()

    def test_iterators_on_empty(self):
        assert P.IntervalDict().values() == []
        assert P.IntervalDict().items() == []
        assert P.IntervalDict().keys() == []
        assert P.IntervalDict().domain() == P.empty()

    def test_combine_empty(self):
        add = lambda x, y: x + y
        assert P.IntervalDict().combine(P.IntervalDict(), add) == P.IntervalDict()

        d = P.IntervalDict([(P.closed(0, 3), 0)])
        assert P.IntervalDict().combine(d, add) == d
        assert d.combine(P.IntervalDict(), add) == d

    def test_combine_nonempty(self):
        add = lambda x, y: x + y
        d1 = P.IntervalDict([(P.closed(1, 3) | P.closed(5, 7), 1)])
        d2 = P.IntervalDict([(P.closed(2, 4) | P.closed(6, 8), 2)])
        assert d1.combine(d2, add) == d2.combine(d1, add)
        assert d1.combine(d2, add) == P.IntervalDict([
            (P.closedopen(1, 2) | P.closedopen(5, 6), 1),
            (P.closed(2, 3) | P.closed(6, 7), 3),
            (P.openclosed(3, 4) | P.openclosed(7, 8), 2),
        ])

        d1 = P.IntervalDict({
            P.closed(0, 1): 2,
            P.closed(3, 4): 2
        })
        d2 = P.IntervalDict({
            P.closed(1, 3): 3,
            P.closed(4, 5): 1
        })
        assert d1.combine(d2, add) == d2.combine(d1, add)
        assert d1.combine(d2, add) == P.IntervalDict({
            P.closedopen(0, 1): 2,
            P.singleton(1): 5,
            P.open(1, 3): 3,
            P.singleton(3): 5,
            P.open(3, 4): 2,
            P.singleton(4): 3,
            P.openclosed(4, 5): 1,
        })

    def test_containment(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        assert 0 in d
        assert -1 not in d
        assert P.closed(-2, -1) not in d
        assert P.closed(1, 2) in d
        assert P.closed(1, 4) not in d

    def test_repr(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        assert repr(d) == '{' + repr(P.closed(0, 3)) + ': 0}'

    def test_pop_value(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        t = d.pop(2)
        assert t == 0
        assert d.get(2, None) is None

    def test_pop_missing_value(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        with pytest.raises(KeyError):
            d.pop(4)

        t = d.pop(4, 1)
        assert t == 1

    def test_pop_interval(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        t = d.pop(P.closed(0, 1))
        assert t.items() == [(P.closed(0, 1), 0)]
        assert d.items() == [(P.openclosed(1, 3), 0)]

        t = d.pop(P.closed(0, 2), 1)
        assert t.items() == [(P.closed(0, 1), 1), (P.openclosed(1, 2), 0)]
        assert d.items() == [(P.openclosed(2, 3), 0)]

    def test_popitem(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        t = d.popitem()
        assert t.items() == [(P.closed(0, 3), 0)]
        assert len(d) == 0

    def test_popitem_with_empty(self):
        with pytest.raises(KeyError):
            P.IntervalDict().popitem()

    def test_clear(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        d.clear()
        assert d == P.IntervalDict()
        d.clear()
        assert d == P.IntervalDict()

    def test_find(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        assert d.find(-1) == P.empty()
        assert d.find(0) == P.closed(0, 3)

    def test_find_on_unions(self):
        d = P.IntervalDict([(P.closed(0, 2), 0), (P.closed(3, 5), 0), (P.closed(7, 9), 1)])
        assert d.find(1) == P.closed(7, 9)
        assert d.find(0) == P.closed(0, 2) | P.closed(3, 5)

    def test_copy(self):
        d = P.IntervalDict([(P.closed(0, 3), 0)])
        assert d.copy() == d
        assert d.copy() is not d

    def test_copy_and_update(self):
        d = P.IntervalDict({P.closed(0, 2): 0, P.closed(4, 5): 1})
        assert d == P.IntervalDict([(P.closed(0, 2), 0), (P.closed(4, 5), 1)])

        a, b = d.copy(), d.copy()
        a.update({P.closed(-1, 1): 2})
        b.update([[P.closed(-1, 1), 2]])
        assert a != d
        assert a == b
        assert a != 1
        assert a.items() == [(P.closed(-1, 1), 2), (P.openclosed(1, 2), 0), (P.closed(4, 5), 1)]

        assert P.IntervalDict([(0, 0), (1, 1)]) == P.IntervalDict([(1, 1), (0, 0)])



import pytest

import portion as P


class TestToString:
    def test_bounds(self):
        assert P.to_string(P.closed(0, 1)) == '[0,1]'
        assert P.to_string(P.openclosed(0, 1)) == '(0,1]'
        assert P.to_string(P.closedopen(0, 1)) == '[0,1)'
        assert P.to_string(P.open(0, 1)) == '(0,1)'

    def test_singleton(self):
        assert P.to_string(P.singleton(0)) == '[0]'

    def test_empty(self):
        assert P.to_string(P.empty()) == '()'

    def test_infinities(self):
        assert P.to_string(P.openclosed(-P.inf, 1)) == '(-inf,1]'
        assert P.to_string(P.closedopen(1, P.inf)) == '[1,+inf)'

    def test_unions(self):
        assert P.to_string(P.closed(0, 1) | P.closed(2, 3)) == '[0,1] | [2,3]'

    def test_bound_types(self):
        assert P.to_string(P.closed('a', 'b')) == "['a','b']"
        assert P.to_string(P.closed(tuple([0]), tuple([1]))) == '[(0,),(1,)]'

    def test_parameters(self):
        i1, i2, i3, i4 = P.closed(0, 1), P.openclosed(0, 1), P.closedopen(0, 1), P.open(0, 1)
        params = {
            'disj': ' or ',
            'sep': '-',
            'left_open': '<!',
            'left_closed': '<',
            'right_open': '!>',
            'right_closed': '>',
            'conv': lambda s: '"{}"'.format(s),
            'pinf': '+oo',
            'ninf': '-oo',
        }

        assert P.to_string(i1, **params) == '<"0"-"1">'
        assert P.to_string(i2, **params) == '<!"0"-"1">'
        assert P.to_string(i3, **params) == '<"0"-"1"!>'
        assert P.to_string(i4, **params) == '<!"0"-"1"!>'

        assert P.to_string(P.empty(), **params) == '<!!>'
        assert P.to_string(P.singleton(1), **params) == '<"1">'

        assert P.to_string(P.openclosed(-P.inf, 1), **params) == '<!-oo-"1">'
        assert P.to_string(P.closedopen(1, P.inf), **params) == '<"1"-+oo!>'

        assert P.to_string(P.closed(0, 1) | P.closed(2, 3), **params) == '<"0"-"1"> or <"2"-"3">'


class TestFromString:
    def test_bounds(self):
        assert P.from_string('[0,1]', int) == P.closed(0, 1)
        assert P.from_string('(0,1]', int) == P.openclosed(0, 1)
        assert P.from_string('[0,1)', int) == P.closedopen(0, 1)
        assert P.from_string('(0,1)', int) == P.open(0, 1)

    def test_singleton(self):
        assert P.from_string('[0]', int) == P.singleton(0)

    def test_empty(self):
        assert P.from_string('()', int) == P.empty()

    def test_infinities(self):
        assert P.from_string('(-inf,1]', int) == P.openclosed(-P.inf, 1)
        assert P.from_string('[1,+inf)', int) == P.closedopen(1, P.inf)

    def test_unions(self):
        assert P.from_string('[0,1] | [2,3]', int) == P.closed(0, 1) | P.closed(2, 3)

    def test_conv_is_required(self):
        with pytest.raises(Exception):
            P.from_string('[1,2]', None)

    def test_parameters(self):
        i1, i2, i3, i4 = '<"0"-"1">', '<!"0"-"1">', '<"0"-"1"!>', '<!"0"-"1"!>'
        params = {
            'conv': lambda s: int(s[1:-1]),
            'disj': ' or ',
            'sep': '-',
            'left_open': '<!',
            'left_closed': '<',
            'right_open': '!>',
            'right_closed': '>',
            'pinf': r'\+oo',
            'ninf': '-oo',
        }

        assert P.from_string(i1, **params) == P.closed(0, 1)
        assert P.from_string(i2, **params) == P.openclosed(0, 1)
        assert P.from_string(i3, **params) == P.closedopen(0, 1)
        assert P.from_string(i4, **params) == P.open(0, 1)

        assert P.from_string('<!!>', **params) == P.empty()
        assert P.from_string('<"1">', **params) == P.singleton(1)

        assert P.from_string('<!-oo-"1">', **params) == P.openclosed(-P.inf, 1)
        assert P.from_string('<"1"-+oo!>', **params) == P.closedopen(1, P.inf)

        assert P.from_string('<"0"-"1"> or <"2"-"3">', **params) == P.closed(0, 1) | P.closed(2, 3)


class TestStringIdentity:
    def test_identity(self):
        i1, i2, i3, i4 = P.closed(0, 1), P.openclosed(0, 1), P.closedopen(0, 1), P.open(0, 1)

        assert P.from_string(P.to_string(i1), int) == i1
        assert P.from_string(P.to_string(i2), int) == i2
        assert P.from_string(P.to_string(i3), int) == i3
        assert P.from_string(P.to_string(i4), int) == i4


class TestToData:
    def test_bounds(self):
        assert P.to_data(P.closed(0, 1)) == [(True, 0, 1, True)]
        assert P.to_data(P.openclosed(0, 1)) == [(False, 0, 1, True)]
        assert P.to_data(P.closedopen(0, 1)) == [(True, 0, 1, False)]
        assert P.to_data(P.open(0, 1)) == [(False, 0, 1, False)]

    def test_values(self):
        assert P.to_data(P.closed('a', 'b')) == [(True, 'a', 'b', True)]
        assert P.to_data(P.closed(tuple([0]), tuple([1]))) == [(True, (0,), (1,), True)]

    def test_singleton(self):
        assert P.to_data(P.singleton(0)) == [(True, 0, 0, True)]

    def test_open_intervals(self):
        assert P.to_data(P.open(-P.inf, P.inf)) == [(False, float('-inf'), float('inf'), False)]
        assert P.to_data(P.openclosed(-P.inf, 0)) == [(False, float('-inf'), 0, True)]
        assert P.to_data(P.closedopen(0, P.inf)) == [(True, 0, float('inf'), False)]

    def test_empty_interval(self):
        assert P.to_data(P.empty()) == [(False, float('inf'), float('-inf'), False)]

    def test_unions(self):
        i = P.openclosed(-P.inf, 4) | P.closedopen(6, P.inf)
        assert P.to_data(i) == [(False, float('-inf'), 4, True), (True, 6, float('inf'), False)]

    def test_parameters(self):
        i = P.openclosed(-P.inf, 4) | P.closedopen(6, P.inf)
        assert P.to_data(i, conv=str, pinf='highest', ninf='lowest') == [(False, 'lowest', '4', True), (True, '6', 'highest', False)]


class TestFromData:
    def test_bounds(self):
        assert P.from_data([(P.CLOSED, 0, 1, P.CLOSED)]) == P.closed(0, 1)
        assert P.from_data([(P.OPEN, 0, 1, P.CLOSED)]) == P.openclosed(0, 1)
        assert P.from_data([(P.CLOSED, 0, 1, P.OPEN)]) == P.closedopen(0, 1)
        assert P.from_data([(P.OPEN, 0, 1, P.OPEN)]) == P.open(0, 1)

    def test_values(self):
        assert P.from_data([(P.CLOSED, 'a', 'b', P.CLOSED)]) == P.closed('a', 'b')
        assert P.from_data([(P.CLOSED, (0,), (1,), P.CLOSED)]) == P.closed(tuple([0]), tuple([1]))

    def test_singleton(self):
        assert P.from_data([(P.CLOSED, 0, 0, P.CLOSED)]) == P.singleton(0)

    def test_open_intervals(self):
        assert P.from_data([(P.OPEN, float('-inf'), float('inf'), P.OPEN)]) == P.open(-P.inf, P.inf)
        assert P.from_data([(P.OPEN, float('-inf'), 0, P.CLOSED)]) == P.openclosed(-P.inf, 0)
        assert P.from_data([(P.CLOSED, 0, float('inf'), P.OPEN)]) == P.closedopen(0, P.inf)

    def test_empty_interval(self):
        assert P.from_data([(P.OPEN, float('inf'), float('-inf'), P.OPEN)]) == P.empty()

    def test_unions(self):
        d = [(P.OPEN, float('-inf'), 4, P.CLOSED), (P.CLOSED, 6, float('inf'), P.OPEN)]
        assert P.from_data(d) == P.openclosed(-P.inf, 4) | P.closedopen(6, P.inf)

    def test_parameters(self):
        d = [(P.OPEN, 'lowest', '4', P.CLOSED), (P.CLOSED, '6', 'highest', P.OPEN)]
        assert P.from_data(d, conv=int, pinf='highest', ninf='lowest') == P.openclosed(-P.inf, 4) | P.closedopen(6, P.inf)


class TestDataIdentity:
    def test_identity(self):
        i1, i2, i3, i4 = P.closed(0, 1), P.openclosed(0, 1), P.closedopen(0, 1), P.open(0, 1)

        assert P.from_data(P.to_data(i2)) == i2
        assert P.from_data(P.to_data(i3)) == i3
        assert P.from_data(P.to_data(i4)) == i4
        assert P.from_data(P.to_data(i1)) == i1



import pytest

import portion as P


class TestIterate:
    def test_default_parameters(self):
        assert list(P.iterate(P.closed(0, 2), step=1)) == [0, 1, 2]
        assert list(P.iterate(P.closedopen(0, 2), step=1)) == [0, 1]
        assert list(P.iterate(P.openclosed(0, 2), step=1)) == [1, 2]
        assert list(P.iterate(P.open(0, 2), step=1)) == [1]
        assert list(P.iterate(P.open(0, 2.5), step=1)) == [1, 2]

    def test_empty_intervals(self):
        assert list(P.iterate(P.empty(), step=1)) == []
        assert list(P.iterate(P.open(0, 1), step=1)) == []

    def test_open_intervals(self):
        with pytest.raises(ValueError):
            list(P.iterate(P.openclosed(-P.inf, 2), step=1))

        gen = P.iterate(P.closedopen(0, P.inf), step=1)
        assert next(gen) == 0
        assert next(gen) == 1
        assert next(gen) == 2  # and so on

    def test_with_gaps(self):
        assert list(P.iterate(P.closed(0, 1) | P.closed(5, 6), step=1)) == [0, 1, 5, 6]
        assert list(P.iterate(P.closed(0, 1) | P.closed(2.5, 4), step=1)) == [0, 1, 2.5, 3.5]
        assert list(P.iterate(P.open(0, 1) | P.open(1, 2), step=1)) == []
        assert list(P.iterate(P.open(0.5, 1) | P.open(1, 3), step=1)) == [2]

    def test_with_step(self):
        assert list(P.iterate(P.closed(0, 6), step=2)) == [0, 2, 4, 6]
        assert list(P.iterate(P.closed(0, 6), step=4)) == [0, 4]
        assert list(P.iterate(P.closed(0, 6), step=lambda x: x + 2)) == [0, 2, 4, 6]

    def test_with_base(self):
        assert list(P.iterate(P.closed(0.4, 2), step=1, base=lambda x: round(x))) == [1, 2]
        assert list(P.iterate(P.closed(0.6, 2), step=1, base=lambda x: round(x))) == [1, 2]

    def test_reversed_iteration(self):
        assert list(P.iterate(P.closed(0, 1), step=-1, reverse=True)) == [1, 0]
        assert list(P.iterate(P.open(0, 3), step=-1, reverse=True)) == [2, 1]
        assert list(P.iterate(P.closed(0, 1), step=-0.5, reverse=True)) == [1, 0.5, 0]
        assert list(P.iterate(P.closed(0, 2), step=-1, base=lambda x: x-1, reverse=True)) == [1, 0]
        assert list(P.iterate(P.closed(0, 2) | P.closed(4, 5), step=-1, reverse=True)) == [5, 4, 2, 1, 0]

    def test_reversed_iteration_with_open_intervals(self):
        with pytest.raises(ValueError):
            list(P.iterate(P.closedopen(0, P.inf), step=-1, reverse=True))

        gen = P.iterate(P.openclosed(-P.inf, 0), step=-1, reverse=True)
        assert next(gen) == 0
        assert next(gen) == -1
        assert next(gen) == -2  # and so on



from collections import namedtuple
from .const import Bound, inf


Atomic = namedtuple('Atomic', ['left', 'lower', 'upper', 'right'])


def mergeable(a, b):
    """
    Tester whether two atomic intervals can be merged (i.e. they overlap or are adjacent).

    :param a: an atomic interval.
    :param b: an atomic interval.
    :return: True if mergeable, False otherwise.
    """
    if a.lower < b.lower or (a.lower == b.lower and a.left == Bound.CLOSED):
        first, second = a, b
    else:
        first, second = b, a

    if first.upper == second.lower:
        return first.right == Bound.CLOSED or second.left == Bound.CLOSED

    return first.upper > second.lower


def open(lower, upper):
    """
    Create an open interval with given bounds.

    :param lower: value of the lower bound.
    :param upper: value of the upper bound.
    :return: an interval.
    """
    return Interval.from_atomic(Bound.OPEN, lower, upper, Bound.OPEN)


def closed(lower, upper):
    """
    Create a closed interval with given bounds.

    :param lower: value of the lower bound.
    :param upper: value of the upper bound.
    :return: an interval.
    """
    return Interval.from_atomic(Bound.CLOSED, lower, upper, Bound.CLOSED)


def openclosed(lower, upper):
    """
    Create a left-open interval with given bounds.

    :param lower: value of the lower bound.
    :param upper: value of the upper bound.
    :return: an interval.
    """
    return Interval.from_atomic(Bound.OPEN, lower, upper, Bound.CLOSED)


def closedopen(lower, upper):
    """
    Create a right-open interval with given bounds.

    :param lower: value of the lower bound.
    :param upper: value of the upper bound.
    :return: an interval.
    """
    return Interval.from_atomic(Bound.CLOSED, lower, upper, Bound.OPEN)


def singleton(value):
    """
    Create a singleton interval.

    :param value: value of the lower and upper bounds.
    :return: an interval.
    """
    return Interval.from_atomic(Bound.CLOSED, value, value, Bound.CLOSED)


def empty():
    """
    Create an empty interval.

    :return: an interval.
    """
    return Interval.from_atomic(Bound.OPEN, inf, -inf, Bound.OPEN)


class Interval:
    """
    This class represents an interval.

    An interval is an (automatically simplified) union of atomic intervals.
    It can be created with Interval.from_atomic(), by passing intervals to __init__, or by using
    one of the helpers provided in this module (open, closed, openclosed, etc.)
    """

    __slots__ = ('_intervals',)

    def __init__(self, *intervals):
        """
        Create an interval from a list of intervals.

        :param intervals: a list of intervals.
        """
        self._intervals = list()

        for interval in intervals:
            if isinstance(interval, Interval):
                if not interval.empty:
                    self._intervals.extend(interval._intervals)
            else:
                raise TypeError('Parameters must be Interval instances')

        if len(self._intervals) == 0:
            # So we have at least one (empty) interval
            self._intervals.append(Atomic(Bound.OPEN, inf, -inf, Bound.OPEN))
        else:
            # Sort intervals by lower bound
            self._intervals.sort(key=lambda i: i.lower)

            i = 0
            # Attempt to merge consecutive intervals
            while i < len(self._intervals) - 1:
                current = self._intervals[i]
                successor = self._intervals[i + 1]

                if mergeable(current, successor):
                    if current.lower == successor.lower:
                        lower = current.lower
                        left = current.left if current.left == Bound.CLOSED else successor.left
                    else:
                        lower = min(current.lower, successor.lower)
                        left = current.left if lower == current.lower else successor.left

                    if current.upper == successor.upper:
                        upper = current.upper
                        right = current.right if current.right == Bound.CLOSED else successor.right
                    else:
                        upper = max(current.upper, successor.upper)
                        right = current.right if upper == current.upper else successor.right

                    union = Atomic(left, lower, upper, right)
                    self._intervals.pop(i)  # pop current
                    self._intervals.pop(i)  # pop successor
                    self._intervals.insert(i, union)
                else:
                    i = i + 1

    @property
    def left(self):
        """
        Lowest left boundary is either CLOSED or OPEN.
        """
        return self._intervals[0].left

    @property
    def lower(self):
        """
        Lowest lower bound value.
        """
        return self._intervals[0].lower

    @property
    def upper(self):
        """
        Highest upper bound value.
        """
        return self._intervals[-1].upper

    @property
    def right(self):
        """
        Highest right boundary is either CLOSED or OPEN.
        """
        return self._intervals[-1].right

    @property
    def empty(self):
        """
        True if interval is empty, False otherwise.
        """
        return (
            self.lower > self.upper or
            (self.lower == self.upper and (self.left == Bound.OPEN or self.right == Bound.OPEN))
        )

    @property
    def atomic(self):
        """
        True if this interval is atomic, False otherwise.
        An interval is atomic if it is composed of a single (possibly empty) atomic interval.
        """
        return len(self._intervals) == 1

    @staticmethod
    def from_atomic(left, lower, upper, right):
        """
        Create an Interval instance containing a single atomic interval.

        :param left: either CLOSED or OPEN.
        :param lower: value of the lower bound.
        :param upper: value of the upper bound.
        :param right: either CLOSED or OPEN.
        """
        instance = Interval()
        left = left if lower not in [inf, -inf] else Bound.OPEN
        right = right if upper not in [inf, -inf] else Bound.OPEN

        instance._intervals = [Atomic(left, lower, upper, right)]
        if instance.empty:
            return Interval()

        return instance

    @property
    def enclosure(self):
        """
        Return the smallest interval composed of a single atomic interval that encloses
        the current interval.

        :return: an Interval instance.
        """
        return Interval.from_atomic(self.left, self.lower, self.upper, self.right)

    def replace(self, left=None, lower=None, upper=None, right=None, *, ignore_inf=True):
        """
        Create a new interval based on the current one and the provided values.

        If current interval is not atomic, it is extended or restricted such that
        its enclosure satisfies the new bounds. In other words, its new enclosure
        will be equal to self.enclosure.replace(left, lower, upper, right).

        Callable can be passed instead of values. In that case, it is called with the current
        corresponding value except if ignore_inf if set (default) and the corresponding
        bound is an infinity.

        :param left: (a function of) left boundary.
        :param lower: (a function of) value of the lower bound.
        :param upper: (a function of) value of the upper bound.
        :param right: (a function of) right boundary.
        :param ignore_inf: ignore infinities if functions are provided (default is True).
        :return: an Interval instance
        """
        enclosure = self.enclosure

        if callable(left):
            left = left(enclosure.left)
        else:
            left = enclosure.left if left is None else left

        if callable(lower):
            if ignore_inf and enclosure.lower in [-inf, inf]:
                lower = enclosure.lower
            else:
                lower = lower(enclosure.lower)
        else:
            lower = enclosure.lower if lower is None else lower

        if callable(upper):
            if ignore_inf and enclosure.upper in [-inf, inf]:
                upper = enclosure.upper
            else:
                upper = upper(enclosure.upper)
        else:
            upper = enclosure.upper if upper is None else upper

        if callable(right):
            right = right(enclosure.right)
        else:
            right = enclosure.right if right is None else right

        if self.atomic:
            return Interval.from_atomic(left, lower, upper, right)

        n_interval = self & Interval.from_atomic(left, lower, upper, right)

        if n_interval.atomic:
            return n_interval.replace(left, lower, upper, right)
        else:
            lowest = n_interval[0].replace(left=left, lower=lower)
            highest = n_interval[-1].replace(upper=upper, right=right)
            return Interval(*[lowest] + n_interval[1:-1] + [highest])

    def apply(self, func):
        """
        Apply a function on each of the underlying atomic intervals and return their union
        as a new interval instance

        Given function is expected to return an interval (possibly empty or not atomic) or
        a 4-uple (left, lower, upper, right) whose values correspond to the parameters of
        Interval.from_atomic(left, lower, upper, right).

        This method is merely a shortcut for Interval(*list(map(func, self))).

        :param func: function to apply on each underlying atomic interval.
        :return: an Interval instance.
        """
        intervals = []

        for i in self:
            value = func(i)

            if isinstance(value, Interval):
                intervals.append(value)
            elif isinstance(value, tuple):
                intervals.append(Interval.from_atomic(*value))
            else:
                raise TypeError('Unsupported return type {} for {}'.format(type(value), value))

        return Interval(*intervals)

    def adjacent(self, other):
        """
        Test if given interval is adjacent.

        An interval is adjacent if there is no intersection, and their union is an atomic interval.

        :param other: an interval.
        :return: True if intervals are adjacent, False otherwise.
        """
        return (self & other).empty and (self | other).atomic

    def overlaps(self, other):
        """
        Test if intervals have any overlapping value.

        :param other: an interval.
        :return: True if intervals overlap, False otherwise.
        """
        if isinstance(other, Interval):
            i_iter = iter(self)
            o_iter = iter(other)
            i_current = next(i_iter)
            o_current = next(o_iter)

            while i_current is not None and o_current is not None:
                if i_current < o_current:
                    i_current = next(i_iter, None)
                elif o_current < i_current:
                    o_current = next(o_iter, None)
                else:
                    return True
            return False
        else:
            raise TypeError('Unsupported type {} for {}'.format(type(other), other))

    def intersection(self, other):
        """
        Return the intersection of two intervals.

        :param other: an interval.
        :return: the intersection of the intervals.
        """
        return self & other

    def union(self, other):
        """
        Return the union of two intervals.

        :param other: an interval.
        :return: the union of the intervals.
        """
        return self | other

    def contains(self, item):
        """
        Test if given item is contained in this interval.
        This method accepts intervals and arbitrary comparable values.

        :param item: an interval or any arbitrary comparable value.
        :return: True if given item is contained, False otherwise.
        """
        return item in self

    def complement(self):
        """
        Return the complement of this interval.

        :return: the complement of this interval.
        """
        return ~self

    def difference(self, other):
        """
        Return the difference of two intervals.

        :param other: an interval.
        :return: the difference of the intervals.
        """
        return self - other

    def __len__(self):
        return len(self._intervals)

    def __iter__(self):
        return iter([Interval.from_atomic(*i) for i in self._intervals])

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [Interval.from_atomic(*i) for i in self._intervals[item]]
        else:
            return Interval.from_atomic(*self._intervals[item])

    def __and__(self, other):
        if not isinstance(other, Interval):
            return NotImplemented

        if self.atomic and other.atomic:
            if self.lower == other.lower:
                lower = self.lower
                left = self.left if self.left == Bound.OPEN else other.left
            else:
                lower = max(self.lower, other.lower)
                left = self.left if lower == self.lower else other.left

            if self.upper == other.upper:
                upper = self.upper
                right = self.right if self.right == Bound.OPEN else other.right
            else:
                upper = min(self.upper, other.upper)
                right = self.right if upper == self.upper else other.right

            return Interval.from_atomic(left, lower, upper, right)
        else:
            intersections = []

            i_iter = iter(self)
            o_iter = iter(other)
            i_current = next(i_iter)
            o_current = next(o_iter)

            while i_current is not None and o_current is not None:
                if i_current < o_current:
                    i_current = next(i_iter, None)
                elif o_current < i_current:
                    o_current = next(o_iter, None)
                else:
                    # i_current and o_current have an overlap
                    intersections.append(i_current & o_current)

                    if i_current <= o_current:
                        # o_current can still intersect next i
                        i_current = next(i_iter, None)
                    elif o_current <= i_current:
                        # i_current can still intersect next o
                        o_current = next(o_iter, None)
                    else:
                        assert False

            return Interval(*intersections)

    def __or__(self, other):
        if isinstance(other, Interval):
            return Interval(self, other)
        else:
            return NotImplemented

    def __contains__(self, item):
        if isinstance(item, Interval):
            if self.atomic:
                left = item.lower > self.lower or (
                    item.lower == self.lower and
                    (item.left == self.left or self.left == Bound.CLOSED)
                )
                right = item.upper < self.upper or (
                    item.upper == self.upper and
                    (item.right == self.right or self.right == Bound.CLOSED)
                )
                return left and right
            else:
                selfiter = iter(self)
                current = next(selfiter)

                for other in item:
                    while current < other:
                        current = next(selfiter)
                    # here current and other could have an overlap
                    if other not in current:
                        return False
                return True
        else:
            # Item is a value
            for i in self._intervals:
                left = (item >= i.lower) if i.left == Bound.CLOSED else (item > i.lower)
                right = (item <= i.upper) if i.right == Bound.CLOSED else (item < i.upper)
                if left and right:
                    return True
            return False

    def __invert__(self):
        complements = [
            Interval.from_atomic(Bound.OPEN, -inf, self.lower, ~self.left),
            Interval.from_atomic(~self.right, self.upper, inf, Bound.OPEN)
        ]

        for i, j in zip(self._intervals[:-1], self._intervals[1:]):
            complements.append(
                Interval.from_atomic(~i.right, i.upper, j.lower, ~j.left)
            )

        return Interval(*complements)

    def __sub__(self, other):
        if isinstance(other, Interval):
            return self & ~other
        else:
            return NotImplemented

    def __eq__(self, other):
        if isinstance(other, Interval):
            if len(other._intervals) != len(self._intervals):
                return False

            for a, b in zip(self._intervals, other._intervals):
                eq = (
                    a.left == b.left and
                    a.lower == b.lower and
                    a.upper == b.upper and
                    a.right == b.right
                )
                if not eq:
                    return False
            return True
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Interval):
            if self.right == Bound.OPEN:
                return self.upper <= other.lower
            else:
                return self.upper < other.lower or \
                    (self.upper == other.lower and other.left == Bound.OPEN)
        else:
            return self.upper < other or (self.right == Bound.OPEN and self.upper == other)

    def __gt__(self, other):
        if isinstance(other, Interval):
            if self.left == Bound.OPEN:
                return self.lower >= other.upper
            else:
                return self.lower > other.upper or \
                    (self.lower == other.upper and other.right == Bound.OPEN)
        else:
            return self.lower > other or (self.left == Bound.OPEN and self.lower == other)

    def __le__(self, other):
        if isinstance(other, Interval):
            if self.right == Bound.OPEN:
                return self.upper <= other.upper
            else:
                return self.upper < other.upper or \
                    (self.upper == other.upper and other.right == Bound.CLOSED)
        else:
            return self.lower < other or (self.left == Bound.CLOSED and self.lower == other)

    def __ge__(self, other):
        if isinstance(other, Interval):
            if self.left == Bound.OPEN:
                return self.lower >= other.lower
            else:
                return self.lower > other.lower or \
                    (self.lower == other.lower and other.left == Bound.CLOSED)
        else:
            return self.upper > other or (self.right == Bound.CLOSED and self.upper == other)

    def __hash__(self):
        return hash(tuple([self.lower, self.upper]))

    def __repr__(self):
        if self.empty:
            return '()'
        elif self.lower == self.upper:
            return '[{}]'.format(repr(self.lower))
        else:
            return ' | '.join(
                '{}{},{}{}'.format(
                    '[' if i.left == Bound.CLOSED else '(',
                    repr(i.lower),
                    repr(i.upper),
                    ']' if i.right == Bound.CLOSED else ')',
                )
                for i in self._intervals
            )



import pytest

import portion as P


class TestHelpers:
    def test_bounds(self):
        assert P.closed(0, 1) == P.Interval.from_atomic(P.CLOSED, 0, 1, P.CLOSED)
        assert P.open(0, 1) == P.Interval.from_atomic(P.OPEN, 0, 1, P.OPEN)
        assert P.openclosed(0, 1) == P.Interval.from_atomic(P.OPEN, 0, 1, P.CLOSED)
        assert P.closedopen(0, 1) == P.Interval.from_atomic(P.CLOSED, 0, 1, P.OPEN)
        assert P.singleton(2) == P.closed(2, 2)

    def test_with_infinities(self):
        assert P.closed(-P.inf, P.inf) == P.open(-P.inf, P.inf)
        assert P.closed(-P.inf, 0) == P.openclosed(-P.inf, 0)
        assert P.closed(0, P.inf) == P.closedopen(0, P.inf)

    def test_empty(self):
        assert P.empty() == P.Interval.from_atomic(P.OPEN, P.inf, -P.inf, P.open)
        assert P.closed(3, -3) == P.empty()

        assert P.openclosed(0, 0) == P.empty()
        assert P.closedopen(0, 0) == P.empty()
        assert P.open(0, 0) == P.empty()
        assert P.closed(0, 0) != P.empty()

        assert P.singleton(P.inf) == P.empty()
        assert P.singleton(-P.inf) == P.empty()


class TestInterval:
    def test_creation(self):
        assert P.Interval() == P.empty()
        assert P.Interval(P.closed(0, 1)) == P.closed(0, 1)
        assert P.Interval(P.closed(0, 1)) == P.closed(0, 1)
        assert P.Interval(P.closed(0, 1), P.closed(2, 3)) == P.closed(0, 1) | P.closed(2, 3)
        assert P.Interval(P.closed(0, 1) | P.closed(2, 3)) == P.closed(0, 1) | P.closed(2, 3)

        with pytest.raises(TypeError):
            P.Interval(1)

    def test_creation_issue_19(self):
        # https://github.com/AlexandreDecan/python-intervals/issues/19
        assert P.Interval(P.empty(), P.empty()) == P.empty()

    def test_bounds(self):
        i = P.openclosed(1, 2)
        assert i.left == P.OPEN
        assert i.right == P.CLOSED
        assert i.lower == 1
        assert i.upper == 2

    def test_bounds_on_empty(self):
        i = P.empty()
        assert i.left == P.OPEN
        assert i.right == P.OPEN
        assert i.lower == P.inf
        assert i.upper == -P.inf

        i = P.openclosed(10, -10)
        assert i.left == P.OPEN
        assert i.right == P.OPEN
        assert i.lower == P.inf
        assert i.upper == -P.inf

        i = P.open(0, 1) | P.closed(3, 4)
        assert i.left == P.OPEN
        assert i.right == P.CLOSED
        assert i.lower == 0
        assert i.upper == 4

    def test_bounds_on_union(self):
        i = P.closedopen(0, 1) | P.openclosed(3, 4)
        assert i.left == P.CLOSED
        assert i.right == P.CLOSED
        assert i.lower == 0
        assert i.upper == 4

    def test_is_empty(self):
        assert P.openclosed(1, 1).empty
        assert P.closedopen(1, 1).empty
        assert P.open(1, 1).empty
        assert not P.closed(1, 1).empty
        assert P.Interval().empty
        assert P.empty().empty

    def test_hash_with_hashable(self):
        assert hash(P.closed(0, 1)) is not None
        assert hash(P.closed(0, 1)) != hash(P.closed(1, 2))

        assert hash(P.openclosed(-P.inf, 0)) is not None
        assert hash(P.closedopen(0, P.inf)) is not None
        assert hash(P.empty()) is not None

        assert hash(P.closed(0, 1) | P.closed(3, 4)) is not None
        assert hash(P.closed(0, 1) | P.closed(3, 4)) != hash(P.closed(0, 1))
        assert hash(P.closed(0, 1) | P.closed(3, 4)) != hash(P.closed(3, 4))

    def test_hash_with_unhashable(self):
        # Let's create a comparable but no hashable object
        class T(int):
            def __hash__(self):
                raise TypeError()

        x = P.closed(T(1), T(2))
        with pytest.raises(TypeError):
            hash(x)

        y = x | P.closed(3, 4)
        with pytest.raises(TypeError):
            hash(y)

    def test_enclosure(self):
        assert P.closed(0, 1) == P.closed(0, 1).enclosure
        assert P.open(0, 1) == P.open(0, 1).enclosure
        assert P.closed(0, 4) == (P.closed(0, 1) | P.closed(3, 4)).enclosure
        assert P.openclosed(0, 4) == (P.open(0, 1) | P.closed(3, 4)).enclosure


class TestIntervalReplace:
    def test_replace_bounds(self):
        i = P.open(-P.inf, P.inf)
        assert i.replace(lower=lambda v: 1, upper=lambda v: 1) == P.open(-P.inf, P.inf)
        assert i.replace(lower=lambda v: 1, upper=lambda v: 2, ignore_inf=False) == P.open(1, 2)

    def test_replace_values(self):
        i = P.open(0, 1)
        assert i.replace(left=P.CLOSED, right=P.CLOSED) == P.closed(0, 1)
        assert i.replace(lower=1, upper=2) == P.open(1, 2)
        assert i.replace(lower=lambda v: 1, upper=lambda v: 2) == P.open(1, 2)

    def test_replace_values_on_infinities(self):
        i = P.open(-P.inf, P.inf)
        assert i.replace(lower=lambda v: 1, upper=lambda v: 2) == i
        assert i.replace(lower=lambda v: 1, upper=lambda v: 2, ignore_inf=False) == P.open(1, 2)

    def test_replace_with_union(self):
        i = P.closed(0, 1) | P.open(2, 3)
        assert i.replace() == i
        assert i.replace(P.OPEN, -1, 4, P.OPEN) == P.openclosed(-1, 1) | P.open(2, 4)
        assert i.replace(lower=2) == P.closedopen(2, 3)
        assert i.replace(upper=1) == P.closedopen(0, 1)
        assert i.replace(lower=5) == P.empty()
        assert i.replace(upper=-5) == P.empty()
        assert i.replace(left=lambda v: ~v, lower=lambda v: v - 1, upper=lambda v: v + 1, right=lambda v: ~v) == P.openclosed(-1, 1) | P.openclosed(2, 4)

    def test_replace_with_empty(self):
        assert P.empty().replace(left=P.CLOSED, right=P.CLOSED) == P.empty()
        assert P.empty().replace(lower=1, upper=2) == P.open(1, 2)
        assert P.empty().replace(lower=lambda v: 1, upper=lambda v: 2) == P.empty()
        assert P.empty().replace(lower=lambda v: 1, upper=lambda v: 2, ignore_inf=False) == P.open(1, 2)


class TestIntervalApply:
    def test_apply(self):
        i = P.closed(0, 1)
        assert i.apply(lambda s: s) == i
        assert i.apply(lambda s: (P.OPEN, -1, 2, P.OPEN)) == P.open(-1, 2)
        assert i.apply(lambda s: P.Interval.from_atomic(P.OPEN, -1, 2, P.OPEN)) == P.open(-1, 2)
        assert i.apply(lambda s: P.open(-1, 2)) == P.open(-1, 2)

    def test_apply_on_unions(self):
        i = P.closed(0, 1) | P.closed(2, 3)
        assert i.apply(lambda s: s) == i
        assert i.apply(lambda s: (P.OPEN, -1, 2, P.OPEN)) == P.open(-1, 2)
        assert i.apply(lambda s: (~s.left, s.lower - 1, s.upper - 1, ~s.right)) == P.open(-1, 0) | P.open(1, 2)
        assert i.apply(lambda s: P.Interval.from_atomic(P.OPEN, -1, 2, P.OPEN)) == P.open(-1, 2)
        assert i.apply(lambda s: P.open(-1, 2)) == P.open(-1, 2)

        assert i.apply(lambda s: (s.left, s.lower, s.upper * 2, s.right)) == P.closed(0, 6)

    def test_apply_on_empty(self):
        assert P.empty().apply(lambda s: (P.CLOSED, 1, 2, P.CLOSED)) == P.closed(1, 2)

    def test_apply_with_incorrect_types(self):
        i = P.closed(0, 1)
        with pytest.raises(TypeError):
            i.apply(lambda s: None)

        with pytest.raises(TypeError):
            i.apply(lambda s: 'unsupported')


class TestIntervalAdjacent():
    def test_adjacent(self):
        assert P.closedopen(0, 1).adjacent(P.closedopen(1, 2))
        assert P.closed(0, 1).adjacent(P.open(1, 2))
        assert not P.closed(0, 1).adjacent(P.closed(1, 2))
        assert not P.open(0, 1).adjacent(P.open(1, 2))

    def test_reversed_adjacent(self):
        assert P.closedopen(1, 2).adjacent(P.closedopen(0, 1))
        assert P.open(1, 2).adjacent(P.closed(0, 1))
        assert not P.closed(1, 2).adjacent(P.closed(0, 1))
        assert not P.open(1, 2).adjacent(P.open(0, 1))

    def test_non_adjacent(self):
        assert not P.closedopen(0, 1).adjacent(P.closedopen(3, 4))
        assert not P.closed(0, 1).adjacent(P.open(3, 4))
        assert not P.closed(0, 1).adjacent(P.closed(3, 4))
        assert not P.open(0, 1).adjacent(P.open(3, 4))

        assert not P.closedopen(3, 4).adjacent(P.closedopen(0, 1))
        assert not P.open(3, 4).adjacent(P.closed(0, 1))
        assert not P.closed(3, 4).adjacent(P.closed(0, 1))
        assert not P.open(3, 4).adjacent(P.open(0, 1))

    def test_overlapping(self):
        assert not P.openclosed(0, 2).adjacent(P.closedopen(2, 3))
        assert not P.closed(0, 2).adjacent(P.closedopen(2, 3))
        assert not P.closed(0, 2).adjacent(P.closed(2, 3))
        assert not P.open(0, 2).adjacent(P.open(2, 3))

        assert not P.closedopen(2, 3).adjacent(P.openclosed(0, 2))
        assert not P.closedopen(2, 3).adjacent(P.closed(0, 2))
        assert not P.closed(2, 3).adjacent(P.closed(0, 2))
        assert not P.open(2, 3).adjacent(P.open(0, 2))

    def test_contained(self):
        assert not P.closed(0, 4).adjacent(P.closed(0, 2))
        assert not P.closed(0, 4).adjacent(P.closed(2, 4))
        assert not P.closed(0, 4).adjacent(P.open(0, 2))
        assert not P.closed(0, 4).adjacent(P.open(2, 4))

        assert not P.closed(0, 2).adjacent(P.closed(0, 4))
        assert not P.closed(2, 4).adjacent(P.closed(0, 4))
        assert not P.closed(0, 2).adjacent(P.open(0, 4))
        assert not P.closed(2, 4).adjacent(P.open(0, 4))

        assert not P.closed(0, 2).adjacent(P.closed(0, 2))
        assert not P.open(0, 2).adjacent(P.open(0, 2))
        assert not P.openclosed(0, 2).adjacent(P.openclosed(0, 2))
        assert not P.closedopen(0, 2).adjacent(P.closedopen(0, 2))

    def test_same_bounds(self):
        assert not P.closed(0, 2).adjacent(P.open(0, 2))
        assert not P.open(0, 2).adjacent(P.closed(0, 2))
        assert not P.openclosed(0, 2).adjacent(P.closedopen(0, 2))
        assert not P.closedopen(0, 2).adjacent(P.openclosed(0, 2))

    def test_nonatomic_interval(self):
        assert (P.closed(0, 1) | P.closed(2, 3)).adjacent(P.open(1, 2))
        assert P.open(1, 2).adjacent(P.closed(0, 1) | P.closed(2, 3))
        assert not (P.closed(0, 1) | P.closed(2, 3)).adjacent(P.closed(1, 2))
        assert (P.closedopen(0, 1) | P.openclosed(2, 3)).adjacent(P.open(-1, 0) | P.closed(1, 2) | P.openclosed(3, 4))


class TestIntervalOverlaps():
    def test_overlaps(self):
        assert P.closed(1, 2).overlaps(P.closed(2, 3))
        assert P.closed(1, 2).overlaps(P.closedopen(2, 3))
        assert P.openclosed(1, 2).overlaps(P.closed(2, 3))
        assert P.openclosed(1, 2).overlaps(P.closedopen(2, 3))

    def test_overlaps_with_nonoverlaping(self):
        assert not P.closed(0, 1).overlaps(P.closed(3, 4))
        assert not P.closed(3, 4).overlaps(P.closed(0, 1))

    def test_overlaps_with_edge_cases(self):
        assert not P.closed(0, 1).overlaps(P.open(1, 2))
        assert not P.closed(0, 1).overlaps(P.openclosed(1, 2))
        assert not P.closedopen(0, 1).overlaps(P.closed(1, 2))
        assert not P.closedopen(0, 1).overlaps(P.closedopen(1, 2))
        assert not P.closedopen(0, 1).overlaps(P.openclosed(1, 2))
        assert not P.closedopen(0, 1).overlaps(P.open(1, 2))
        assert not P.open(0, 1).overlaps(P.open(1, 2))
        assert P.open(0, 2).overlaps(P.open(0, 1))
        assert P.open(0, 1).overlaps(P.open(0, 2))

    def test_overlaps_with_empty(self):
        assert not P.empty().overlaps(P.open(-P.inf, P.inf))
        assert not P.open(-P.inf, P.inf).overlaps(P.empty())

    def test_overlaps_with_itself(self):
        assert P.closed(0, 1).overlaps(P.closed(0, 1))
        assert P.closed(0, 1).overlaps(P.open(0, 1))
        assert P.open(0, 1).overlaps(P.closed(0, 1))
        assert P.closed(0, 1).overlaps(P.openclosed(0, 1))
        assert P.closed(0, 1).overlaps(P.closedopen(0, 1))

    def test_overlaps_with_incompatible_types(self):
        with pytest.raises(TypeError):
            P.closed(0, 1).overlaps(1)


class TestIntervalComparison:
    @pytest.mark.parametrize('i1,i2,i3', [
        (P.closed(0, 1), P.closed(1, 2), P.closed(2, 3)),
        (P.open(0, 2), P.open(1, 3), P.open(2, 4)),
    ])
    def test_equalities(self, i1, i2, i3):
        assert i1 == i1
        assert i1 != i2 and i2 != i1
        assert i1 != i3 and i3 != i1
        assert i2 != i3 and i3 != i2

        assert not i1 == 1

    @pytest.mark.parametrize('i1,i2,i3', [
        (P.closed(0, 1), P.closed(1, 2), P.closed(2, 3)),
        (P.open(0, 2), P.open(1, 3), P.open(2, 4)),
    ])
    def test_inequalities(self, i1, i2, i3):
        assert i1 < i3 and i3 > i1
        assert i1 <= i2 and i2 >= i1
        assert i1 <= i3 and i3 >= i1
        assert not i1 < i2 and not i2 > i1

    def test_closed_atomic_with_values(self):
        i = P.closed(0, 5)

        assert -1 < i
        assert -1 <= i
        assert 6 > i
        assert 6 >= i

        assert not (2 < i)
        assert 2 <= i
        assert not (5 < i)
        assert 5 <= i

        assert not (3 > i)
        assert 3 >= i
        assert not (0 > i)
        assert 0 >= i

    def test_open_atomic_with_values(self):
        i = P.open(0, 5)

        assert -1 < i
        assert -1 <= i
        assert 6 > i
        assert 6 >= i

        assert not (2 < i)
        assert 2 <= i
        assert not (5 < i)
        assert not (5 <= i)

        assert not (3 > i)
        assert 3 >= i
        assert not (0 > i)
        assert not (0 >= i)

    def test_atomic_with_infinities(self):
        i = P.closed(0, 5)
        assert -P.inf < i
        assert -P.inf <= i
        assert not (-P.inf > i)
        assert not (-P.inf >= i)

        assert P.inf > i
        assert P.inf >= i
        assert not (P.inf < i)
        assert not (P.inf <= i)

        i = P.open(0, 5)
        assert -P.inf < i
        assert -P.inf <= i
        assert not (-P.inf > i)
        assert not (-P.inf >= i)

        assert P.inf > i
        assert P.inf >= i
        assert not (P.inf < i)
        assert not (P.inf <= i)

    def test_with_values(self):
        i = P.closedopen(0, 10)

        assert -1 < i
        assert -1 <= i
        assert not (0 < i)
        assert 0 <= i
        assert not (5 < i)
        assert 5 <= i
        assert not (10 < i)
        assert not (10 <= i)
        assert not (12 < i)
        assert not (12 <= i)

        assert 12 > i
        assert 12 >= i
        assert 10 > i
        assert 10 >= i
        assert not (8 > i)
        assert 8 >= i
        assert not (0 > i)
        assert 0 >= i
        assert not (-1 > i)
        assert not (-1 >= i)

    def test_with_infinities(self):
        i = P.closedopen(0, 10)

        assert -P.inf < i
        assert -P.inf <= i
        assert not (-P.inf > i)
        assert not (-P.inf >= i)

        assert P.inf > i
        assert P.inf >= i
        assert not (P.inf < i)
        assert not (P.inf <= i)

    def test_with_intervals(self):
        i1, i2, i3 = P.closed(0, 1), P.closed(1, 2), P.closed(2, 3)

        i4 = i1 | i3

        assert i4 != i2 and i2 != i4
        assert not i4 < i2 and not i2 > i4
        assert not i4 > i2 and not i2 < i4
        assert not i4 <= i2
        assert not i4 >= i2
        assert i2 <= i4
        assert i2 >= i4

        i5 = P.closed(5, 6) | P.closed(7, 8)

        assert i4 != i5 and i5 != i4
        assert i4 < i5 and i5 > i4
        assert not i4 > i5 and not i5 < i4
        assert i4 <= i5
        assert i5 >= i4
        assert not i4 >= i5
        assert not i5 <= i4

    def test_with_empty(self):
        assert P.empty() < P.empty()
        assert P.empty() <= P.empty()
        assert P.empty() > P.empty()
        assert P.empty() >= P.empty()

        assert P.empty() < P.closed(2, 3)
        assert P.empty() <= P.closed(2, 3)
        assert P.empty() > P.closed(2, 3)
        assert P.empty() >= P.closed(2, 3)


class TestIntervalContainment:
    def test_with_values(self):
        assert 1 in P.closed(0, 2)
        assert 1 in P.closed(1, 2)
        assert 1 in P.closed(0, 1)

        assert 1 in P.open(0, 2)
        assert 1 not in P.open(0, 1)
        assert 1 not in P.open(1, 2)

        assert 1 in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)
        assert 5 in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)
        assert 10 in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)

        assert -1 not in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)
        assert 3 not in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)
        assert 7 not in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)
        assert 11 not in P.closed(0, 2) | P.closed(4, 6) | P.closed(8, 10)

    def test_with_infinities(self):
        assert 1 in P.closed(-P.inf, P.inf)
        assert 1 in P.closed(-P.inf, 1)
        assert 1 in P.closed(1, P.inf)
        assert 1 not in P.closed(-P.inf, 0)
        assert 1 not in P.closed(2, P.inf)

        assert P.inf not in P.closed(-P.inf, P.inf)
        assert -P.inf not in P.closed(-P.inf, P.inf)

        assert P.inf not in P.closed(0, 1)

    def test_with_empty(self):
        assert 1 not in P.empty()
        assert P.inf not in P.empty()
        assert -P.inf not in P.empty()

    def test_with_intervals(self):
        assert P.closed(1, 2) in P.closed(0, 3)
        assert P.closed(1, 2) in P.closed(1, 2)
        assert P.open(1, 2) in P.closed(1, 2)
        assert P.closed(1, 2) not in P.open(1, 2)
        assert P.closed(0, 1) not in P.closed(1, 2)
        assert P.closed(0, 2) not in P.closed(1, 3)
        assert P.closed(-P.inf, P.inf) in P.closed(-P.inf, P.inf)
        assert P.closed(0, 1) in P.closed(-P.inf, P.inf)
        assert P.closed(-P.inf, P.inf) not in P.closed(0, 1)

    def test_with_unions(self):
        assert P.closed(0, 1) | P.closed(2, 3) in P.closed(0, 4)
        assert P.closed(0, 1) | P.closed(2, 3) in P.closed(0, 1) | P.closed(2, 3)
        assert P.closed(0, 1) | P.closed(2, 3) in P.closed(0, 0) | P.closed(0, 1) | P.closed(2, 3)

        assert P.closed(0, 1) | P.closed(2, 3) not in P.closed(0, 2)
        assert P.closed(0, 1) | P.closed(2, 3) not in P.closed(0, 1) | P.closedopen(2, 3)
        assert P.closed(0, 1) | P.closed(2, 3) not in P.closed(0, 1) | P.closedopen(2, 3) | P.openclosed(3, 4)

    def test_with_empty_intervals(self):
        assert P.empty() in P.closed(0, 3)
        assert P.empty() in P.empty()
        assert P.closed(0, 0) not in P.empty()
        assert P.singleton(0) | P.singleton(1) not in P.empty()

    def test_proxy_method(self):
        i1, i2 = P.closed(0, 1), P.closed(2, 3)
        assert (1 in i1) == i1.contains(1)
        assert (i1 in i2) == i2.contains(i1)


class TestIntervalIntersection:
    def test_with_itself(self):
        assert P.closed(0, 1) & P.closed(0, 1) == P.closed(0, 1)
        assert P.closed(0, 1) & P.open(0, 1) == P.open(0, 1)
        assert P.openclosed(0, 1) & P.closedopen(0, 1) == P.open(0, 1)

    def test_with_adjacent(self):
        assert P.closed(0, 2) & P.closed(2, 4) == P.singleton(2)
        assert P.open(0, 2) & P.open(2, 4) == P.empty()

    def test_with_containment(self):
        assert P.closed(0, 4) & P.closed(2, 3) == P.closed(2, 3)
        assert P.closed(0, 4) & P.closed(0, 3) == P.closed(0, 3)
        assert P.closed(0, 4) & P.open(2, 3) == P.open(2, 3)
        assert P.open(0, 4) & P.closed(2, 3) == P.closed(2, 3)
        assert P.open(0, 4) & P.closed(3, 4) == P.closedopen(3, 4)

    def test_with_overlap(self):
        assert P.closed(0, 3) & P.closed(2, 4) == P.closed(2, 3)
        assert P.open(0, 3) & P.closed(2, 4) == P.closedopen(2, 3)

    def test_with_union(self):
        assert (P.closed(0, 2) | P.closed(4, 6)) & (P.closed(0, 1) | P.closed(4, 5)) == P.closed(0, 1) | P.closed(4, 5)
        assert (P.closed(0, 2) | P.closed(4, 6)) & (P.closed(-1, 1) | P.closed(3, 6)) == P.closed(0, 1) | P.closed(4, 6)
        assert (P.closed(0, 2) | P.closed(4, 6)) & (P.closed(1, 4) | P.singleton(5)) == P.closed(1, 2) | P.singleton(4) | P.singleton(5)

    def test_empty(self):
        assert (P.closed(0, 1) & P.closed(2, 3)).empty
        assert P.closed(0, 1) & P.empty() == P.empty()

    def test_proxy_method(self):
        i1, i2 = P.closed(0, 1), P.closed(2, 3)
        assert i1 & i2 == i1.intersection(i2)

    def test_with_invalid_type(self):
        with pytest.raises(TypeError):
            P.closed(0, 1) & 1


class TestIntervalUnion:
    def test_atomic(self):
        assert P.closed(1, 2) | P.closed(1, 2) == P.closed(1, 2)
        assert P.closed(1, 4) | P.closed(2, 3) == P.closed(1, 4)
        assert P.closed(1, 2) | P.closed(2, 3) == P.closed(2, 3) | P.closed(1, 2)
        assert P.closed(1, 2) | P.closed(3, 4) == P.closed(1, 2) | P.closed(3, 4)

    def test_with_itself(self):
        assert P.closed(1, 2) | P.closed(1, 2) == P.closed(1, 2)
        assert P.open(1, 2) | P.closed(1, 2) == P.closed(1, 2)
        assert P.closedopen(1, 2) | P.openclosed(1, 2) == P.closed(1, 2)

    def test_with_containment(self):
        assert P.closed(1, 4) | P.closed(2, 3) == P.closed(1, 4)
        assert P.open(1, 4) | P.closed(2, 3) == P.open(1, 4)

    def test_with_overlap(self):
        assert P.closed(1, 3) | P.closed(2, 4) == P.closed(1, 4)

    def test_with_adjacent(self):
        assert P.closed(1, 2) | P.closed(2, 3) == P.closed(1, 3)
        assert P.closed(1, 2) | P.open(2, 3) == P.closedopen(1, 3)
        assert P.open(1, 2) | P.closed(2, 3) == P.openclosed(1, 3)
        assert P.open(1, 3) | P.open(2, 4) == P.open(1, 4)
        assert P.closedopen(1, 2) | P.closed(2, 3) == P.closed(1, 3)
        assert P.open(1, 2) | P.closed(2, 4) == P.openclosed(1, 4)

    def test_with_disjoint(self):
        assert P.closed(1, 2) | P.closed(3, 4) != P.closed(1, 4)
        assert (P.closed(1, 2) | P.closed(3, 4) | P.closed(2, 3)).atomic
        assert P.closed(1, 2) | P.closed(3, 4) | P.closed(2, 3) == P.closed(1, 4)
        assert P.closed(1, 2) | P.closed(0, 4) == P.closed(0, 4)

        assert (P.closed(0, 1) | P.closed(2, 3) | P.closed(1, 2)).atomic
        assert P.closed(0, 1) | P.closed(2, 3) | P.closed(1, 2) == P.closed(0, 3)

    def test_with_empty(self):
        assert P.closed(0, 1) | P.empty() == P.closed(0, 1)

    def test_issue_12(self):
        # https://github.com/AlexandreDecan/python-intervals/issues/12
        assert P.open(0, 2) | P.closed(0, 2) == P.closed(0, 2)
        assert P.open(0, 2) | P.closed(1, 2) == P.openclosed(0, 2)
        assert P.open(0, 2) | P.closed(0, 1) == P.closedopen(0, 2)

        assert P.closed(0, 2) | P.open(0, 2) == P.closed(0, 2)
        assert P.closed(1, 2) | P.open(0, 2) == P.openclosed(0, 2)
        assert P.closed(0, 1) | P.open(0, 2) == P.closedopen(0, 2)

        assert P.closed(0, 2) | P.singleton(2) == P.closed(0, 2)
        assert P.closedopen(0, 2) | P.singleton(2) == P.closed(0, 2)
        assert P.openclosed(0, 2) | P.singleton(2) == P.openclosed(0, 2)
        assert P.openclosed(0, 2) | P.singleton(0) == P.closed(0, 2)

        assert P.singleton(2) | P.closed(0, 2) == P.closed(0, 2)
        assert P.singleton(2) | P.closedopen(0, 2) == P.closed(0, 2)
        assert P.singleton(2) | P.openclosed(0, 2) == P.openclosed(0, 2)
        assert P.singleton(0) | P.openclosed(0, 2) == P.closed(0, 2)

    def test_issue_13(self):
        # https://github.com/AlexandreDecan/python-intervals/issues/13
        assert P.closed(1, 1) | P.openclosed(1, 2) == P.closed(1, 2)
        assert P.openclosed(1, 2) | P.closed(1, 1) == P.closed(1, 2)
        assert P.closed(0, 1) | P.openclosed(1, 2) == P.closed(0, 2)
        assert P.openclosed(1, 2) | P.closed(0, 1) == P.closed(0, 2)

        assert P.openclosed(1, 2) | P.closed(1, 1) == P.closed(1, 2)
        assert P.closed(1, 1) | P.openclosed(1, 2) == P.closed(1, 2)
        assert P.openclosed(1, 2) | P.closed(0, 1) == P.closed(0, 2)
        assert P.closed(0, 1) | P.openclosed(1, 2) == P.closed(0, 2)

    def test_proxy_method(self):
        i1, i2 = P.closed(0, 1), P.closed(2, 3)
        assert i1 | i2 == i1.union(i2)

    def test_with_invalid_type(self):
        with pytest.raises(TypeError):
            P.closed(0, 1) | 1


class TestIntervalComplement:
    def test_singleton(self):
        assert ~P.singleton(0) == P.open(-P.inf, 0) | P.open(0, P.inf)
        assert ~(P.open(-P.inf, 0) | P.open(0, P.inf)) == P.singleton(0)

    def test_nonempty(self):
        assert ~P.closed(1, 2) == P.open(-P.inf, 1) | P.open(2, P.inf)
        assert ~P.open(1, 2) == P.openclosed(-P.inf, 1) | P.closedopen(2, P.inf)

    def test_union(self):
        assert ~(P.singleton(0) | P.singleton(5) | P.singleton(10)) == P.open(-P.inf, 0) | P.open(0, 5) | P.open(5, 10) | P.open(10, P.inf)
        assert ~(P.open(0, 1) | P.closed(2, 3) | P.open(4, 5)) == P.openclosed(-P.inf, 0) | P.closedopen(1, 2) | P.openclosed(3, 4) | P.closedopen(5, P.inf)

    @pytest.mark.parametrize('i', [P.closed(0, 1), P.open(0, 1), P.openclosed(0, 1), P.closedopen(0, 1)])
    def test_identity(self, i):
        for interval in i:
            assert ~(~interval) == interval

    def test_empty(self):
        assert ~P.open(1, 1) == P.open(-P.inf, P.inf)
        assert (~P.closed(-P.inf, P.inf)).empty
        assert ~P.empty() == P.open(-P.inf, P.inf)

    def test_proxy_method(self):
        i1, i2 = P.closed(0, 1), P.closed(2, 3)
        assert ~i1 == i1.complement()
        assert ~i2 == i2.complement()


class TestIntervalDifference:
    @pytest.mark.parametrize('i', [P.closed(0, 1), P.open(0, 1), P.openclosed(0, 1), P.closedopen(0, 1), P.empty(), P.singleton(0)])
    def test_with_itself(self, i):
        assert i - i == P.empty()

    def test_with_disjoint(self):
        assert P.closed(0, 1) - P.closed(2, 3) == P.closed(0, 1)
        assert P.closed(0, 4) - P.empty() == P.closed(0, 4)
        assert P.empty() - P.closed(0, 4) == P.empty()

    def test_with_smaller(self):
        assert P.closed(0, 4) - P.closed(2, 3) == P.closedopen(0, 2) | P.openclosed(3, 4)
        assert P.closed(1, 4) - P.closed(1, 3) == P.openclosed(3, 4)
        assert P.closed(1, 4) - P.closed(2, 4) == P.closedopen(1, 2)
        assert P.closed(0, 4) - P.open(1, 2) == P.closed(0, 1) | P.closed(2, 4)
        assert P.closed(0, 2) - P.open(0, 2) == P.singleton(0) | P.singleton(2)

    def test_with_larger(self):
        assert P.closed(0, 2) - P.closed(0, 4) == P.empty()
        assert P.closed(0, 2) - P.closed(-2, 2) == P.empty()
        assert P.closed(0, 2) - P.closed(-2, 4) == P.empty()
        assert P.open(0, 2) - P.closed(0, 2) == P.empty()

    def test_with_overlap(self):
        assert P.closed(0, 2) - P.closed(1, 3) == P.closedopen(0, 1)
        assert P.closed(0, 2) - P.open(1, 3) == P.closed(0, 1)
        assert P.closed(0, 2) - P.closed(-2, 1) == P.openclosed(1, 2)
        assert P.closed(0, 2) - P.open(-2, 1) == P.closed(1, 2)

    def test_proxy_method(self):
        i1, i2 = P.closed(0, 1), P.closed(2, 3)
        assert i1 - i2 == i1.difference(i2)

    def test_with_invalid_type(self):
        with pytest.raises(TypeError):
            P.closed(0, 1) - 1


class TestIntervalIteration:
    def test_length(self):
        i1 = P.closed(10, 10) | P.closed(5, 6) | P.closed(7, 8) | P.closed(8, 9)
        assert len(i1) == 3

    def test_containment(self):
        i1 = P.closed(10, 10) | P.closed(5, 6) | P.closed(7, 8) | P.closed(8, 9)
        for i in i1:
            assert i in i1

    def test_order(self):
        i1 = P.closed(10, 10) | P.closed(5, 6) | P.closed(7, 8) | P.closed(8, 9)
        assert sorted(i1, key=lambda i: i.lower) == list(i1)
        assert sorted(i1, key=lambda i: i.upper) == list(i1)

    def test_indexes(self):
        i1 = P.closed(10, 10) | P.closed(5, 6) | P.closed(7, 8) | P.closed(8, 9)
        assert i1[0] == P.closed(5, 6)
        assert i1[1] == P.closed(7, 9)
        assert i1[2] == P.closed(10, 10)
        assert i1[-1] == P.closed(10, 10)

    def test_slices(self):
        items = [P.closed(5, 6), P.closed(7, 9), P.singleton(10)]
        interval = P.Interval(*items)
        assert interval[:] == items
        assert interval[:2] == items[:2]
        assert interval[::-1] == items[::-1]
        assert interval[::2] == items[::2]

    def test_missing_index(self):
        i1 = P.closed(10, 10) | P.closed(5, 6) | P.closed(7, 8) | P.closed(8, 9)
        with pytest.raises(IndexError):
            i1[3]

    def test_empty(self):
        assert len(P.empty()) == 1
        assert list(P.empty()) == [P.empty()]
        assert P.empty()[0] == P.empty()



import doctest

import portion as P


def test_readme():
    failure = None

    try:
        doctest.testfile('../README.md', raise_on_error=True, globs={'P': P})
    except doctest.DocTestFailure as e:
        failure = e.example.want, e.got, e.example.source

    if failure:
        # Make pytest display it outside the "except" block, to avoid a noisy traceback
        want, got, example = failure
        assert want.strip() == got.strip(), 'DocTest failure in "{}"'.format(example.strip())
        assert False  # In case .strip() removed something useful



from os import path
from setuptools import setup, find_packages
from codecs import open

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='portion',
    version='2.0.0',
    license='LGPLv3',

    author='Alexandre Decan',
    url='https://github.com/AlexandreDecan/portion',

    description='Python data structure and operations for intervals',
    long_description=long_description,
    long_description_content_type='text/markdown',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering :: Mathematics',

        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        # Typing :: Typed
    ],
    keywords='interval operation range math',

    packages=find_packages(include=['portion']),
    python_requires='~=3.5',

    install_requires=[],
    extras_require={
        'test': ['pytest ~= 5.0.1'],
        'travis': ['coverage ~= 5.0.3', 'coveralls ~= 1.11.1']
    },

    zip_safe=True,
)
